from django.contrib import admin

from .models import Tileset

class TilesetAdmin(admin.ModelAdmin):
    readonly_fields=('size',)
    list_display = ('name', 'layer_name', 'server_url', 'created_by', 'created_at')
    search_fields = ['name']

admin.site.register(Tileset, TilesetAdmin)




# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os
import tempfile

BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'e#i*49t%&&=jbfs64hb8(fj(m8gqicz9h3+!3(y#(9k!uu#sd!'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'djmp',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'djmp.urls'


# Database
# https://docs.djangoproject.com/en/1.6/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.6/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.6/howto/static-files/

STATIC_URL = '/static/'

# Applying 12-factor app pattern
FILE_CACHE_DIRECTORY = os.getenv('FILE_CACHE_DIRECTORY', tempfile.gettempdir())


import time
import json

from django.shortcuts import get_object_or_404, render
from django.http import HttpResponseRedirect, HttpResponse
from django.core.urlresolvers import reverse
from django.views import generic
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.conf import settings

from mapproxy.config.config import load_default_config, load_config
from mapproxy.util.ext.dictspec.validator import validate, ValidationError
from mapproxy.config.loader import ProxyConfiguration, ConfigurationError
from mapproxy.wsgiapp import MapProxyApp

from webtest import TestApp as TestApp_

import os
import yaml
import logging
log = logging.getLogger('mapproxy.config')


from .models import Tileset, MAPPROXY_CACHE_DIR
from .helpers import get_status
from .validator import validate_references, validate_options


class IndexView(generic.ListView):
    template_name = 'djmp/index.html'

    def get_queryset(self):
        return Tileset.objects.all()


class DetailView(generic.DetailView):
    model = Tileset
    template_name = 'djmp/tileset_detail.html'


@login_required
def seed(request, pk):
    tileset = get_object_or_404(Tileset, pk=pk)
    return HttpResponse(json.dumps(tileset.seed()))


@login_required
def tileset_status(request, pk):
    tileset = get_object_or_404(Tileset, pk=pk)
    return HttpResponse(json.dumps(get_status(tileset)))





class TestApp(TestApp_):
    """
    Wraps webtest.TestApp and explicitly converts URLs to strings.
    Behavior changed with webtest from 1.2->1.3.
    """
    def get(self, url, *args, **kw):
        kw['expect_errors'] = True
        return TestApp_.get(self, str(url), *args, **kw)


def simple_name(layer_name):
    layer_name = str(layer_name)

    if ':' in layer_name:
        layer_name = layer_name.split(':')[1]

    return layer_name


def get_mapproxy(layer, seed=False, ignore_warnings=True, renderd=False):
    """Creates a mapproxy config for a given layer-like object.
       Compatible with django-registry and GeoNode.
    """
    bbox = [float(layer.bbox_x0), float(layer.bbox_y0), float(layer.bbox_x1), float(layer.bbox_y1)]

    url = str(layer.service.url)

    layer_name = simple_name(layer.name)

    srs = 'EPSG:4326'
    bbox_srs = 'EPSG:4326'
    grid_srs = 'EPSG:3857'

    if layer.type == 'WARPER':
        url = str(layer.url.replace("maps//wms", "maps/wms"))
        grid_srs = 'EPSG:900913'

    if layer.type == 'WM':
        url = str(layer.url.replace("maps//wms", "maps/wms"))

    default_source = {
             'type': 'wms',
             'coverage': {
              'bbox': bbox,
              'srs': srs,
              'bbox_srs': bbox_srs,
              'supported_srs': ['EPSG:4326', 'EPSG:900913', 'EPSG:3857'],
              },
             'req': {
                'layers': layer_name,
                'url': url,
                'transparent': True,
              },
           }

    if layer.type == 'ESRI:ArcGIS:MapServer' or layer.type == 'ESRI:ArcGIS:ImageServer':
        # blindly replace it with /arcgis/
        url = url.replace("/ArcGIS/rest/", "/arcgis/")
        # same for uppercase
        url = url.replace("/arcgis/rest/", "/arcgis/")
        # and for old versions
        url = url.replace("ArcX/rest/services", "arcx/services")
        # in uppercase or lowercase
        url = url.replace("arcx/rest/services", "arcx/services")

        srs = 'EPSG:3857'
        bbox_srs = 'EPSG:3857'

        default_source = {
                  'type': 'arcgis',
                  'req': {
                     'url': str(layer.service.url).split('?')[0],
                     'grid': 'default_grid',
                     'transparent': True,
                  },
        }

    # A source is the WMS config
    sources = {
      'default_source': default_source
    }

    # A grid is where it will be projects (Mercator in our case)
    grids = {
             'default_grid': {
                 'tile_size': [256, 256],
                 'srs': grid_srs,
                 'origin': 'nw',
                 }
             }

    # A cache that does not store for now. It needs a grid and a source.
    caches = {'default_cache':
              {
               'cache':
               {
                   'type': 'file',
                   'directory_layout': 'tms',
                   'directory': os.path.join(cache_directory,
                                             'mapproxy',
                                             'layer',
                                             '%s' % layer.id,
                                             'map',
                                             'wmts',
                                             layer_name,
                                             'default_grid',
                                             ),
               },
               'grids': ['default_grid'],
               'sources': ['default_source']},
              }

    # The layer is connected to the cache
    layers = [
        {'name': layer_name,
         'sources': ['default_cache'],
         'title': str(layer.title),
         },
    ]

    # Services expose all layers.
    # WMS is used for reprojecting
    # TMS is used for easy tiles
    # Demo is used to test our installation, may be disabled in final version
    services = {
      'wms': {'image_formats': ['image/png'],
              'md': {'abstract': 'This is the Harvard HyperMap Proxy.',
                     'title': 'Harvard HyperMap Proxy'},
              'srs': ['EPSG:4326', 'EPSG:3857'],
              'versions': ['1.1.1']},
      'wmts': {
              'restful': True,
              'restful_template':
              '/{Layer}/{TileMatrixSet}/{TileMatrix}/{TileCol}/{TileRow}.png',
              },
      'tms': {
              'origin': 'nw',
              },
      'demo': None,
    }

    global_config = {
      'http': {'ssl_no_cert_checks': True},
    }

    # Start with a sane configuration using MapProxy's defaults
    conf_options = load_default_config()

    # Populate a dictionary with custom config changes
    extra_config = {
        'caches': caches,
        'grids': grids,
        'layers': layers,
        'services': services,
        'sources': sources,
        'globals': global_config,
    }

    yaml_config = yaml.dump(extra_config, default_flow_style=False)
    # If you want to test the resulting configuration. Turn on the next
    # line and use that to generate a yaml config.
    # assert False

    # Merge both
    load_config(conf_options, config_dict=extra_config)

    # Make sure the config is valid.
    errors, informal_only = validate_options(conf_options)
    for error in errors:
        log.warn(error)
    if not informal_only or (errors and not ignore_warnings):
        raise ConfigurationError('invalid configuration')

    errors = validate_references(conf_options)
    for error in errors:
        log.warn(error)

    conf = ProxyConfiguration(conf_options, seed=seed, renderd=renderd)

    # Create a MapProxy App
    app = MapProxyApp(conf.configured_services(), conf.base_config)

    # Wrap it in an object that allows to get requests by path as a string.
    return TestApp(app), yaml_config



import logging
import helpers
import tempfile

from django.db import models
from django.conf import settings
from django.core.validators import MaxValueValidator, MinValueValidator
from mapproxy.seed.config import SeedConfigurationError, ConfigurationError

MAPPROXY_CACHE_DIR = getattr(settings, 'MAPPROXY_CACHE_DIR', tempfile.gettempdir())

log = logging.getLogger('djmapproxy')

CACHE_TYPES = [
    ['file', 'file'],
    #['mbtiles','mbtiles'],
    #['sqllite', 'sqllite'],
    ['geopackage', 'geopackage']
]

DIR_LAYOUTS = [
    ['tms', 'tms'],
    ['tc', 'TileCache']
]

SERVER_SERVICE_TYPES = [
    ['wms','wms'],
    ['tile', 'tile'],
    ['mapnik', 'mapnik']
]

class Tileset(models.Model):

    # base
    name = models.CharField(unique=True, max_length=30)
    created_by = models.CharField(max_length=256)
    created_at = models.DateTimeField(auto_now_add=True)

    # server
    server_url = models.URLField()
    server_service_type = models.CharField(max_length=10, choices=SERVER_SERVICE_TYPES)
    server_username = models.CharField(blank=True, max_length=30)
    server_password = models.CharField(blank=True, max_length=30)

    # layer
    layer_name = models.CharField(blank=True, max_length=200)
    layer_zoom_start = models.IntegerField(blank=True, default=0)
    layer_zoom_stop = models.IntegerField()

    # area
    bbox_x0 = models.DecimalField(max_digits=19, decimal_places=15, default=-180, validators = [MinValueValidator(-180), MaxValueValidator(180)])
    bbox_x1 = models.DecimalField(max_digits=19, decimal_places=15, default=180, validators = [MinValueValidator(-180), MaxValueValidator(180)])
    bbox_y0 = models.DecimalField(max_digits=19, decimal_places=15, default=-89.9, validators = [MinValueValidator(-89.9), MaxValueValidator(89.9)])
    bbox_y1 = models.DecimalField(max_digits=19, decimal_places=15, default=89.9, validators = [MinValueValidator(-89.9), MaxValueValidator(89.9)])

    # cache
    cache_type = models.CharField(max_length=10, choices=CACHE_TYPES)
    # file cache params
    directory_layout = models.CharField(max_length=20, choices=DIR_LAYOUTS, blank=True, null=True)
    directory = models.CharField(max_length=256, default=MAPPROXY_CACHE_DIR, blank=True, null=True)
    # gpkg cache params
    filename = models.CharField(max_length=256, blank=True, null=True)
    table_name = models.CharField(max_length=128, blank=True, null=True)

    # size
    size = models.CharField('Size (MB)', default='0', max_length=128)

    def __unicode__(self):
        return self.name

    # terminate the seeding of this tileset!
    def stop(self):
        log.debug('tileset.stop')
        res = {'status': 'not in progress'}
        pid_str = helpers.get_pid_from_lock_file(self)
        process = helpers.get_process_from_pid(pid_str)
        if process:
            log.debug('tileset.stop, will stop, pid: {}').format(pid_str)
            res = {'status': 'stopped'}
            children = process.children()
            for c in children:
                c.terminate()
            process.terminate()
        else:
            if pid_str == 'preparing_to_start':
                res = {'status': 'debug, prevent start!'}
                # TODO: prevent it from starting!
                log.debug('process not running but may be started shortly')
            elif helpers.is_int_str(pid_str):
                log.debug('tileset.stop, process not running but cleaned lock file')

        helpers.remove_lock_file(self.id)
        return res

    def seed(self):
        
        lock_file = helpers.get_lock_file(self)
        if lock_file:
            log.debug('generating tileset')
            try:
                pid = helpers.seed_process_spawn(self)
                lock_file.write("{}\n".format(pid))
                res = {'status': 'started'}
            except (SeedConfigurationError, ConfigurationError) as e:
                log.error('Something went wrong when generating.. removing lock file')

                helpers.remove_lock_file(self)
                res = {'status': 'unable to start',
                       'error': e.message}
            finally:
                lock_file.flush()
                lock_file.close()
        else:
            log.debug('tileset.generate, will NOT generate. already running, pid: {}'.format(helpers.get_pid_from_lock_file(self)))
            res = {'status': 'already started'}

        return res



import yaml
import os
import errno
import time
import multiprocessing
import psutil
import logging
from datetime import datetime
from dateutil import parser

from django.utils.text import slugify
from mapproxy.seed.seeder import seed
from mapproxy.seed.config import SeedingConfiguration, SeedConfigurationError, ConfigurationError
from mapproxy.seed.spec import validate_seed_conf
from mapproxy.config.loader import ProxyConfiguration
from mapproxy.config.spec import validate_options
from mapproxy.seed import seeder
from mapproxy.seed import util

from .mapproxy_config import get_mapproxy_conf, get_seed_conf, u_to_str


log = logging.getLogger('djmapproxy')


def generate_confs(tileset, ignore_warnings=True, renderd=False):
    """
    Takes a Tileset object and returns mapproxy and seed config files
    """

    mapproxy_conf_json = get_mapproxy_conf(tileset)
    mapproxy_conf = yaml.safe_load(mapproxy_conf_json)

    seed_conf_json = get_seed_conf(tileset)
    seed_conf = yaml.safe_load(seed_conf_json)

    errors, informal_only = validate_options(mapproxy_conf)
    if not informal_only or (errors and not ignore_warnings):
        raise ConfigurationError('invalid configuration - {}'.format(', '.join(errors)))

    cf = ProxyConfiguration(mapproxy_conf, seed=seed, renderd=renderd)

    errors, informal_only = validate_seed_conf(seed_conf)
    if not informal_only:
        raise SeedConfigurationError('invalid seed configuration - {}'.format(', '.join(errors)))
    seed_cf = SeedingConfiguration(seed_conf, mapproxy_conf=cf)

    return cf, seed_cf


def get_tileset_dir(tileset):
    folder = get_tileset_base_folder(tileset)
    if not os.path.exists(folder):
        os.makedirs(folder)
    return folder


def get_tileset_base_folder(tileset):
    return os.path.join(tileset.directory, tileset.name)


def get_tileset_location(tileset):
    if tileset.cache_type == 'file':
        return get_tileset_dir(tileset)
    elif tileset.cache_type == 'gpkg':
        return '{}/{}.gpkg'.format(get_tileset_dir(tileset), tileset.name)


def get_lock_filename(tileset):
    return '{}/generate_tileset_{}.lck'.format(get_tileset_dir(tileset), tileset.id)


def update_tileset_stats(tileset):
    size, updated = get_tileset_stats(tileset)

    tileset.created_at = updated
    tileset.size = size
    tileset.save()


def is_int_str(v):
    v = str(v).strip()
    return v == '0' or (v if v.find('..') > -1 else v.lstrip('-+').rstrip('0').rstrip('.')).isdigit()


def get_tileset_stats(tileset):
    tileset_location = get_tileset_location(tileset)
    
    stat = os.stat(tileset_location)
    if stat:
        size =  os.popen('du -sh %s' % tileset_location).read().split('\t')[0]
        updated = datetime.fromtimestamp(stat.st_ctime).isoformat()
        return size, updated
    return None


def add_tileset_file_attribs(target_object, tileset):
    size, updated = get_tileset_stats(tileset)
    target_object['size'] = size
    target_object['updated'] = updated


def get_status(tileset):
    res = {
        'current': {
            'status': 'unknown'
        },
        'pending': {
            'status': 'not in progress'
        }
    }

    # generate status for already existing tileset
    # if there is a .gpkg file on disk, get the size and time last updated
    tileset_location = get_tileset_location(tileset)
    if os.path.exists(tileset_location):
        res['current']['status'] = 'ready'
        # get the size and time last updated for the tileset
        add_tileset_file_attribs(res['current'], tileset)
        # get the size and time last updated for the 'pending' tileset
        add_tileset_file_attribs(res['pending'], tileset)

        pid = get_pid_from_lock_file(tileset)

        if pid:
            process = get_is_process_running(pid)
            if process:
                # if tileset generation is in progress
                res['pending']['status'] = 'in progress'
                tileset_location = get_tileset_location(tileset, 'progress_log')
                log_filename = '%s/%s.progress_log' % (tileset_location, tileset.name)
                if os.path.isfile(log_filename):
                    with open(log_filename, 'r') as f:
                        lines = f.read().replace('\r', '\n')
                        lines = lines.split('\n')

                    # an actual progress step update which looks like:
                    # "[15:11:11]  4  50.00% 0.00000, 672645.84891, 18432942.24503, 18831637.78456 (112 tiles) ETA: 2015-07-07-15:11:12"\n
                    latest_step = None
                    # a progress update on the current step which looks like:
                    # "[15:11:16]  87.50%   0000                 ETA: 2015-07-07-15:11:17'\r
                    latest_progress = None
                    if len(lines) > 0:
                        for line in lines[::-1]:
                            tokens = line.split()
                            if len(tokens) > 2:
                                if is_int_str(tokens[1]):
                                    latest_step = tokens
                                    break
                                elif tokens[1].endswith('%'):
                                    if latest_progress is None:
                                        # keep going, don't break
                                        latest_progress = tokens
                                        continue
                    if latest_step:
                        # if we have a step %, up date the progress %
                        if latest_progress:
                            latest_step[2] = latest_progress[1]

                        res['pending']['progress'] = latest_step[2][0:-1]
                        res['pending']['current_zoom_level'] = latest_step[1]

                        # get the eta but pass if date is cannot be parsed.
                        try:
                            iso_date = parser.parse(latest_step[len(latest_step) - 1]).isoformat()
                            res['pending']['estimated_completion_time'] = iso_date
                        except ValueError:
                            pass
                else:
                    res['pending']['status'] = 'in progress, but log not found'
            else:
                res['pending']['status'] = 'stopped'
    else:
        res.pop('pending', None)
        res['current']['status'] = 'not generated'

    
    return res
    

# when using uwsgi, several processes each with their own interpreter are often launched. This means that the typical
# multiprocessing sync mechanisms such as Lock and Manager cannot be used. comments about issues to know about uwsgi:
# http://uwsgi-docs.readthedocs.org/en/latest/ThingsToKnow.html note that enable-threads, close-on-exec, and
# close-on-exec2 were not effective and even if they were, other deployments will need to match uwsgi setting which is
# inconvenient especially since the problems caused can be misleading. The implementation here uses lock files to check
# if a tileset is being generated and which process to kill when generate needs to be stopped by a user. Using celery
# for multiprocessing poses another problem: since it generates its worker pool processes as proc.daemon = True,
# each celery process cannot invoke the mapproxy.seed function which in turn wants to launch other processes. This
# can be fixed in celery but it requires a patch. celery project reopened this 'bug' a few days ago as of 7-22-2015:
# https://github.com/celery/celery/issues/1709 if it is fixed, we can switch to using celery without any immediate
# gain for the current use case. Instead of using daemon processes, they should use another mechanism to track/kill
# child processes so that each celery task can launch other processes.
def seed_process_spawn(tileset):
    mapproxy_conf, seed_conf = generate_confs(tileset)

    # if there is an old _generating one around, back it up
    tileset_location = get_tileset_location(tileset)

    backup_millis = int(round(time.time() * 1000))
    generating_filename = '%s/%s.generating' % (tileset_location, tileset.name)
    if os.path.isfile(generating_filename):
        os.rename(generating_filename, '{}_{}'.format(generating_filename, backup_millis))

    # if there is an old progress_log around, back it up
    log_filename = '%s/%s.progress_log' % (tileset_location, tileset.name)
    if os.path.isfile(log_filename):
        os.rename(log_filename, '{}_{}'.format(log_filename, backup_millis))

    # generate the new gpkg as name.generating file
    out = open(log_filename, 'w+')
    progress_logger = util.ProgressLog(out=out, verbose=True, silent=False)
    tasks = seed_conf.seeds(['tileset_seed'])
    # launch the task using another process
    process = multiprocessing.Process(target=seed_process_target, args=(tileset, tasks, progress_logger))
    pid = None
    if 'preparing_to_start' == get_pid_from_lock_file(tileset):
        process.start()
        pid = process.pid
    else:
        log.debug(' Not starting process. cancel was requested.')
    return pid


def seed_process_target(tileset, tasks, progress_logger):
    seeder.seed(tasks=tasks, progress_logger=progress_logger)
    log.debug('start seeding. tileset {}'.format(tileset.id))
    # now that we have generated the new gpkg file, backup the last one, then rename
    # the _generating one to the main name
    # if os.path.isfile(get_tileset_filename(tileset_name)):
    #     millis = int(round(time.time() * 1000))
    #     os.rename(get_tileset_filename(tileset_name), '{}_{}'.format(get_tileset_filename(tileset_name), millis))
    # os.rename(get_tileset_filename(tileset_name, 'generating'), get_tileset_filename(tileset_name))
    remove_lock_file(tileset)


def get_lock_file(tileset):
    flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY
    lock_file = None

    try:
        file_handle = os.open(get_lock_filename(tileset), flags)
    except OSError as e:
        if e.errno == errno.EEXIST:
            # Failed, file already exists.
            pass
        else:
            # Something unexpected went wrong so re-raise the exception.
            raise
    else:
        # No exception, so the file must have been created successfully.
        lock_file = os.fdopen(file_handle, 'w')
        lock_file.write('preparing_to_start\n')
        lock_file.flush()
    return lock_file


def remove_lock_file(tileset):
    try:
        os.remove(get_lock_filename(tileset))
    except OSError as e:
        # Error removing lock file
        log.debug('There was a problem removing the lock file. Something: {}'.format(e.errno))
        pass


def get_pid_from_lock_file(tileset):
    pid = None
    name = get_lock_filename(tileset)
    if os.path.isfile(name):
        with open(name, 'r') as lock_file:
            lines = lock_file.readlines()
            if len(lines) > 0:
                if lines[-1]:
                    pid = lines[-1].rstrip()
    return pid


def get_process_from_pid(pid):
    process = None
    if is_int_str(pid):
        try:
            process = psutil.Process(pid=int(pid))
        except (psutil.NoSuchProcess, psutil.ZombieProcess):
            log.debug('PROCESS HAS BEEN TERMINATED {}'.format(int(pid)))
            pass
    return process


def get_is_process_running(pid):
    process = None
    if is_int_str(pid):
        try:
            process = psutil.Process(pid=int(pid))
            exitCode = process.wait(0)
            log.debug('waited for process.. exitCode: {}'.format(exitCode))
        except (psutil.NoSuchProcess, psutil.ZombieProcess, psutil.TimeoutExpired):
            pass
    return process



from django.test import TestCase
from django.test.client import Client
from django.core.urlresolvers import reverse

from .views import tileset_status, seed
from .models import Tileset


class DjmpTest(TestCase):

    fixtures = ['test_data.json']

    def setUp(self):
        self.user = 'admin'
        self.passwd = 'admin'
        self.client = Client()

    def test_seeding(self):
        """ test seeding"""
        self.client.login(username='admin', password='admin')
        resp = self.client.get(reverse('tileset_seed', args=(1,)))
        self.assertEqual('{"status": "started"}', resp.content)


from django.conf import settings
from django.conf.urls import include, patterns, url
from django.contrib import admin

from .views import DetailView, seed, tileset_status

admin.autodiscover()

urlpatterns = patterns('',

    url(r'^(?P<pk>\d+)/$', DetailView.as_view(), name='tileset_detail'),
    url(r'^(?P<pk>\d+)/seed$', seed, name='tileset_seed'),
    url(r'^(?P<pk>\d+)/status$', tileset_status, name='tileset_status'),
    (r'^admin/', include(admin.site.urls)),
)



# This file is part of the MapProxy project.
# Copyright (C) 2015 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import datetime
import os.path
from mapproxy.compat import string_type, iteritems
from mapproxy.util.ext.dictspec.validator import validate, ValidationError
from mapproxy.util.ext.dictspec.spec import one_of, anything, number
from mapproxy.util.ext.dictspec.spec import recursive, required, type_spec, combined
from mapproxy.compat import string_type

import logging
log = logging.getLogger('mapproxy.config')


TAGGED_SOURCE_TYPES = [
    'wms',
    'mapserver',
    'mapnik'
]


def validate_references(conf_dict):
    validator = Validator(conf_dict)
    return validator.validate()


grids = dict(
    GLOBAL_GEODETIC=dict(
        srs='EPSG:4326', origin='sw', name='GLOBAL_GEODETIC'
    ),
    GLOBAL_MERCATOR=dict(
        srs='EPSG:900913', origin='sw', name='GLOBAL_MERCATOR'
    ),
    GLOBAL_WEBMERCATOR=dict(
        srs='EPSG:3857', origin='nw', name='GLOBAL_WEBMERCATOR'
    )
)

class Validator(object):

    def __init__(self, conf_dict):
        self.sources_conf = conf_dict.get('sources', {})
        self.caches_conf = conf_dict.get('caches', {})
        self.layers_conf = conf_dict.get('layers')
        self.services_conf = conf_dict.get('services')
        self.grids_conf = conf_dict.get('grids')
        self.globals_conf = conf_dict.get('globals')

        self.errors = []
        self.known_grids = set(grids.keys())
        if self.grids_conf:
            self.known_grids.update(self.grids_conf.keys())

    def validate(self):
        if not self.layers_conf:
            self.errors.append("Missing layers section")
        if isinstance(self.layers_conf, dict):
            return []
        if not self.services_conf:
            self.errors.append("Missing services section")

        if len(self.errors) > 0:
            return self.errors

        for layer in self.layers_conf:
            self._validate_layer(layer)

        return self.errors

    def _validate_layer(self, layer):
        layer_sources = layer.get('sources', [])
        tile_sources = layer.get('tile_sources', [])
        child_layers = layer.get('layers', [])

        if not layer_sources and not child_layers and not tile_sources:
            self.errors.append(
                "Missing sources for layer '%s'" % layer.get('name')
            )
        for child_layer in child_layers:
            self._validate_layer(child_layer)

        for source in layer_sources:
            if source in self.caches_conf:
                self._validate_cache(source, self.caches_conf[source])
                continue
            if source in self.sources_conf:
                source, layers = self._split_tagged_source(source)
                self._validate_source(source, self.sources_conf[source], layers)
                continue

            self.errors.append(
                "Source '%s' for layer '%s' not in cache or source section" % (
                    source,
                    layer['name']
                )
            )

        for source in tile_sources:
            if source in self.caches_conf:
                self._validate_cache(source, self.caches_conf[source])
                continue

            self.errors.append(
                "Tile source '%s' for layer '%s' not in cache section" % (
                    source,
                    layer['name']
                )
            )


    def _split_tagged_source(self, source_name):
        layers = None
        if ':' in str(source_name):
            source_name, layers = str(source_name).split(':')
            layers = layers.split(',') if layers is not None else None
        return source_name, layers

    def _validate_source(self, name, source, layers):
        source_type = source.get('type')
        if source_type == 'wms':
            self._validate_wms_source(name, source, layers)
        if source_type == 'mapserver':
            self._validate_mapserver_source(name, source, layers)
        if source_type == 'mapnik':
            self._validate_mapnik_source(name, source, layers)

    def _validate_wms_source(self, name, source, layers):
        if source['req'].get('layers') is None and layers is None:
            self.errors.append("Missing 'layers' for source '%s'" % (
                name
            ))
        if source['req'].get('layers') is not None and layers is not None:
            self._validate_tagged_layer_source(
                name,
                source['req'].get('layers'),
                layers
            )

    def _validate_mapserver_source(self, name, source, layers):
        mapserver = source.get('mapserver')
        if mapserver is None:
            if (
                not self.globals_conf or
                not self.globals_conf.get('mapserver') or
                not self.globals_conf['mapserver'].get('binary')
            ):
                self.errors.append("Missing mapserver binary for source '%s'" % (
                    name
                ))
            elif not os.path.isfile(self.globals_conf['mapserver']['binary']):
                self.errors.append("Could not find mapserver binary (%s)" % (
                    self.globals_conf['mapserver'].get('binary')
                ))
        elif mapserver is None or not source['mapserver'].get('binary'):
            self.errors.append("Missing mapserver binary for source '%s'" % (
                name
            ))
        elif not os.path.isfile(source['mapserver']['binary']):
            self.errors.append("Could not find mapserver binary (%s)" % (
                source['mapserver']['binary']
            ))

        if source['req'].get('layers') and layers is not None:
            self._validate_tagged_layer_source(
                name,
                source['req'].get('layers'),
                layers
            )

    def _validate_mapnik_source(self, name, source, layers):
        if source.get('layers') and layers is not None:
            self._validate_tagged_layer_source(name, source.get('layers'), layers)

    def _validate_tagged_layer_source(self, name, supported_layers, requested_layers):
        if isinstance(supported_layers, string_type):
            supported_layers = [supported_layers]
        if not set(requested_layers).issubset(set(supported_layers)):
            self.errors.append(
                "Supported layers for source '%s' are '%s' but tagged source requested "
                "layers '%s'" % (
                    name,
                    ', '.join(supported_layers),
                    ', '.join(requested_layers)
                ))

    def _validate_cache(self, name, cache):
        if isinstance(cache.get('sources', []), dict):
            self._validate_bands(name, set(cache['sources'].keys()))
            for band, confs  in iteritems(cache['sources']):
                for conf in confs:
                    band_source = conf['source']
                    self._validate_cache_source(name, band_source)
        else:
            for cache_source in cache.get('sources', []):
                self._validate_cache_source(name, cache_source)

        for grid in cache.get('grids', []):
            if grid not in self.known_grids:
                self.errors.append(
                    "Grid '%s' for cache '%s' not found in config" % (
                        grid,
                        name
                    )
                )

    def _validate_cache_source(self, cache_name, source_name):
        source_name, layers = self._split_tagged_source(source_name)
        if self.sources_conf and source_name in self.sources_conf:
            source = self.sources_conf.get(source_name)
            if (
                layers is not None and
                source.get('type') not in TAGGED_SOURCE_TYPES
            ):
                self.errors.append(
                    "Found tagged source '%s' in cache '%s' but tagged sources only "
                    "supported for '%s' sources" % (
                        source_name,
                        cache_name,
                        ', '.join(TAGGED_SOURCE_TYPES)
                    )
                )
                return
            self._validate_source(source_name, source, layers)
            return
        if self.caches_conf and source_name in self.caches_conf:
            self._validate_cache(source_name, self.caches_conf[source_name])
            return
        self.errors.append(
            "Source '%s' for cache '%s' not found in config" % (
                source_name,
                cache_name
            )
        )

    def _validate_bands(self, cache_name, bands):
        if 'l' in bands and len(bands) > 1:
            self.errors.append(
                "Cannot combine 'l' band with bands in cache '%s'" % (
                    cache_name
                )
            )


def validate_options(conf_dict):
    """
    Validate `conf_dict` agains mapproxy.yaml spec.
    Returns tuple with a list of errors and a bool.
    The list is empty when no errors where found.
    The bool is True when the errors are informal and not critical.
    """
    try:
        validate(mapproxy_yaml_spec, conf_dict)
    except ValidationError as ex:
        return ex.errors, ex.informal_only
    else:
        return [], True

coverage = {
    'polygons': str(),
    'polygons_srs': str(),
    'bbox': one_of(str(), [number()]),
    'bbox_srs': str(),
    'ogr_datasource': str(),
    'ogr_where': str(),
    'ogr_srs': str(),
    'datasource': one_of(str(), [number()]),
    'where': str(),
    'srs': str(),
}
image_opts = {
    'mode': str(),
    'colors': number(),
    'transparent': bool(),
    'resampling_method': str(),
    'format': str(),
    'encoding_options': {
        anything(): anything()
    },
    'merge_method': str(),
}

http_opts = {
    'method': str(),
    'client_timeout': number(),
    'ssl_no_cert_checks': bool(),
    'ssl_ca_certs': str(),
    'headers': {
        anything(): str()
    },
}

mapserver_opts = {
    'binary': str(),
    'working_dir': str(),
}

scale_hints = {
    'max_scale': number(),
    'min_scale': number(),
    'max_res': number(),
    'min_res': number(),
}

source_commons = combined(
    scale_hints,
    {
        'concurrent_requests': int(),
        'coverage': coverage,
        'seed_only': bool(),
    }
)

riak_node = {
    'host': str(),
    'pb_port': number(),
    'http_port': number(),
}

cache_types = {
    'file': {
        'directory_layout': str(),
        'use_grid_names': bool(),
        'directory': str(),
        'tile_lock_dir': str(),
    },
    'sqlite': {
        'directory': str(),
        'tile_lock_dir': str(),
    },
    'mbtiles': {
        'filename': str(),
        'tile_lock_dir': str(),
    },
    'couchdb': {
        'url': str(),
        'db_name': str(),
        'tile_metadata': {
            anything(): anything()
        },
        'tile_id': str(),
        'tile_lock_dir': str(),
    },
    'riak': {
        'nodes': [riak_node],
        'protocol': one_of('pbc', 'http', 'https'),
        'bucket': str(),
        'default_ports': {
            'pb': number(),
            'http': number(),
        },
        'secondary_index': bool(),
    }
}

on_error = {
    anything(): {
        required('response'): one_of([int], str),
        'cache': bool,
    }
}



inspire_md = {
    'linked': {
        required('metadata_url'): {
            required('url'): str,
            required('media_type'): str,
        },
        required('languages'): {
            required('default'): str,
        },
    },
    'embedded': {
        required('resource_locators'): [{
            required('url'): str,
            required('media_type'): str,
        }],
        required('temporal_reference'): {
            'date_of_publication': one_of(str, datetime.date),
            'date_of_creation': one_of(str, datetime.date),
            'date_of_last_revision': one_of(str, datetime.date),
        },
        required('conformities'): [{
            'title': string_type,
            'uris': [str],
            'date_of_publication': one_of(str, datetime.date),
            'date_of_creation': one_of(str, datetime.date),
            'date_of_last_revision': one_of(str, datetime.date),
            required('resource_locators'): [{
                required('url'): str,
                required('media_type'): str,
            }],
            required('degree'): str,
        }],
        required('metadata_points_of_contact'): [{
            'organisation_name': string_type,
            'email': str,
        }],
        required('mandatory_keywords'): [str],
        'keywords': [{
            required('title'): string_type,
            'date_of_publication': one_of(str, datetime.date),
            'date_of_creation': one_of(str, datetime.date),
            'date_of_last_revision': one_of(str, datetime.date),
            'uris': [str],
            'resource_locators': [{
                required('url'): str,
                required('media_type'): str,
            }],
            required('keyword_value'): string_type,
        }],
        required('metadata_date'): one_of(str, datetime.date),
        'metadata_url': {
            required('url'): str,
            required('media_type'): str,
        },
        required('languages'): {
            required('default'): str,
        },
    },
}

wms_130_layer_md = {
    'abstract': string_type,
    'keyword_list': [
        {
            'vocabulary': string_type,
            'keywords': [string_type],
        }
    ],
    'attribution': {
        'title': string_type,
        'url':    str,
        'logo': {
            'url':    str,
            'width':  int,
            'height': int,
            'format': string_type,
       }
    },
    'identifier': [
        {
            'url': str,
            'name': string_type,
            'value': string_type,
        }
    ],
    'metadata': [
        {
            'url': str,
            'type': str,
            'format': str,
        },
    ],
    'data': [
        {
            'url': str,
            'format': str,
        }

    ],
    'feature_list': [
        {
            'url': str,
            'format': str,
        }
    ],
}

grid_opts = {
    'base': str(),
    'name': str(),
    'srs': str(),
    'bbox': one_of(str(), [number()]),
    'bbox_srs': str(),
    'num_levels': int(),
    'res': [number()],
    'res_factor': one_of(number(), str()),
    'max_res': number(),
    'min_res': number(),
    'stretch_factor': number(),
    'max_shrink_factor': number(),
    'align_resolutions_with': str(),
    'origin': str(),
    'tile_size': [int()],
    'threshold_res': [number()],
}

ogc_service_md = {
    'title': string_type,
    'abstract': string_type,
    'online_resource': string_type,
    'contact': anything(),
    'fees': string_type,
    'access_constraints': string_type,
    'keyword_list': [
        {
            'vocabulary': string_type,
            'keywords': [string_type],
        }
    ],
}

band_source = {
    required('source'): str(),
    required('band'): int,
    'factor': number(),
}

band_sources = {
    'r': [band_source],
    'g': [band_source],
    'b': [band_source],
    'a': [band_source],
    'l': [band_source],
}

mapproxy_yaml_spec = {
    '__config_files__': anything(), # only used internaly
    'globals': {
        'image': {
            'resampling_method': 'method',
            'paletted': bool(),
            'stretch_factor': number(),
            'max_shrink_factor': number(),
            'jpeg_quality': number(),
            'formats': {
                anything(): image_opts,
            },
            'font_dir': str(),
            'merge_method': str(),
        },
        'http': combined(
            http_opts,
            {
                'access_control_allow_origin': one_of(str(), {}),
            }
        ),
        'cache': {
            'base_dir': str(),
            'lock_dir': str(),
            'tile_lock_dir': str(),
            'meta_size': [number()],
            'meta_buffer': number(),
            'max_tile_limit': number(),
            'minimize_meta_requests': bool(),
            'concurrent_tile_creators': int(),
            'link_single_color_images': bool(),
        },
        'grid': {
            'tile_size': [int()],
        },
        'srs': {
          'axis_order_ne': [str()],
          'axis_order_en': [str()],
          'proj_data_dir': str(),
        },
        'tiles': {
            'expires_hours': number(),
        },
        'mapserver': mapserver_opts,
        'renderd': {
            'address': str(),
        }
    },
    'grids': {
        anything(): grid_opts,
    },
    'caches': {
        anything(): {
            required('sources'): one_of([string_type], band_sources),
            'name': str(),
            'grids': [str()],
            'cache_dir': str(),
            'meta_size': [number()],
            'meta_buffer': number(),
            'minimize_meta_requests': bool(),
            'concurrent_tile_creators': int(),
            'disable_storage': bool(),
            'format': str(),
            'image': image_opts,
            'request_format': str(),
            'use_direct_from_level': number(),
            'use_direct_from_res': number(),
            'link_single_color_images': bool(),
            'watermark': {
                'text': string_type,
                'font_size': number(),
                'color': one_of(str(), [number()]),
                'opacity': number(),
                'spacing': str(),
            },
            'cache': type_spec('type', cache_types)
        }
    },
    'services': {
        'demo': {},
        'kml': {
            'use_grid_names': bool(),
        },
        'tms': {
            'use_grid_names': bool(),
            'origin': str(),
        },
        'wmts': {
            'kvp': bool(),
            'restful': bool(),
            'restful_template': str(),
            'md': ogc_service_md,
        },
        'wms': {
            'srs': [str()],
            'bbox_srs': [one_of(str(), {'bbox': [number()], 'srs': str()})],
            'image_formats': [str()],
            'attribution': {
                'text': string_type,
            },
            'featureinfo_types': [str()],
            'featureinfo_xslt': {
                anything(): str()
            },
            'on_source_errors': str(),
            'max_output_pixels': one_of(number(), [number()]),
            'strict': bool(),
            'md': ogc_service_md,
            'inspire_md': type_spec('type', inspire_md),
            'versions': [str()],
        },
    },

    'sources': {
        anything(): type_spec('type', {
            'wms': combined(source_commons, {
                'wms_opts': {
                    'version': str(),
                    'map': bool(),
                    'featureinfo': bool(),
                    'legendgraphic': bool(),
                    'legendurl': str(),
                    'featureinfo_format': str(),
                    'featureinfo_xslt': str(),
                },
                'image': combined(image_opts, {
                    'opacity':number(),
                    'transparent_color': one_of(str(), [number()]),
                    'transparent_color_tolerance': number(),
                }),
                'supported_formats': [str()],
                'supported_srs': [str()],
                'http': http_opts,
                'forward_req_params': [str()],
                required('req'): {
                    required('url'): str(),
                    anything(): anything()
                }
            }),
            'mapserver': combined(source_commons, {
                    'wms_opts': {
                        'version': str(),
                        'map': bool(),
                        'featureinfo': bool(),
                        'legendgraphic': bool(),
                        'legendurl': str(),
                        'featureinfo_format': str(),
                        'featureinfo_xslt': str(),
                    },
                    'image': combined(image_opts, {
                        'opacity':number(),
                        'transparent_color': one_of(str(), [number()]),
                        'transparent_color_tolerance': number(),
                    }),
                    'supported_formats': [str()],
                    'supported_srs': [str()],
                    'forward_req_params': [str()],
                    required('req'): {
                        required('map'): str(),
                        anything(): anything()
                    },
                    'mapserver': mapserver_opts,
            }),
            'tile': combined(source_commons, {
                required('url'): str(),
                'transparent': bool(),
                'image': image_opts,
                'grid': str(),
                'request_format': str(),
                'origin': str(), # TODO: remove with 1.5
                'http': http_opts,
                'on_error': on_error,
            }),
            'mapnik': combined(source_commons, {
                required('mapfile'): str(),
                'transparent': bool(),
                'image': image_opts,
                'layers': one_of(str(), [str()]),
                'use_mapnik2': bool(),
                'scale_factor': number(),
            }),
            'arcgis': combined(source_commons, {
               required('req'): {
                    required('url'): str(),
                    'dpi': int(),
                    'layers': str(),
                    'transparent': bool(),
                    'time': str()
                },
                'supported_srs': [str()],
                'http': http_opts
            }),
            'debug': {
            },
        })
    },

    'layers': one_of(
        {
            anything(): combined(scale_hints, {
                'sources': [string_type],
                required('title'): string_type,
                'legendurl': str(),
                'md': wms_130_layer_md,
            })
        },
        recursive([combined(scale_hints, {
            'sources': [string_type],
            'tile_sources': [string_type],
            'name': str(),
            required('title'): string_type,
            'legendurl': str(),
            'layers': recursive(),
            'md': wms_130_layer_md,
            'dimensions': {
                anything(): {
                    required('values'): [one_of(string_type, float, int)],
                    'default': one_of(string_type, float, int),
                }
            }
        })])
    ),
     # `parts` can be used for partial configurations that are referenced
     # from other sections (e.g. coverages, dimensions, etc.)
    'parts': anything(),
}



import os
from distutils.core import setup

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(
    name="djmp",
    version="0.2.7",
    author="",
    author_email="",
    description="django-mapproxy adaptor",
    long_description=('djmp'),
    # Full list of classifiers can be found at:
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 1 - Planning',
    ],
    license="BSD",
    keywords="django mapproxy",
    url='https://github.com/terranodo/django-mapproxy',
    packages=['djmp',],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'Django>=1.8.7',
        'MapProxy>=1.9.0',
        'PyYAML',
        'django-tastypie',
        'Shapely',
        'psutil',
        'pyproj',
    ]
)



import json
import os
import base64

from pyproj import Proj, transform

services_conf = {
    "wms":{
        "on_source_errors":"raise",
        "image_formats": ["image/png"]
    }
}

layers_conf = [{
    "name": "",
    "title": "",
    "sources":[
        "tileset_cache"
    ]
}]

caches_conf = {
    "tileset_cache":{
        "grids":[
            "EPSG3857"
        ],
        "sources":[
            "tileset_source"
        ],
        "cache": {}
    }
}

sources_conf = {
    "tileset_source":{
        "type": ""
    }
}

grids_conf = {
    "EPSG3857": {
        "origin": "nw",
        "srs": "EPSG:3857"
    }
}


seed_coverages_conf = {
    "tileset_geom": {
        "bbox": [],
        "srs": "EPSG:3857"
    }
}

seed_seeds_conf = {
    "tileset_seed": {
        "refresh_before": {
            "minutes": 0
        },
        "caches": [
            "tileset_cache"
        ],
        "levels": {
            "from": 0,
            "to": 12
        },
        "coverages": ["tileset_geom"]
    }
}


def get_mapproxy_conf(tileset):

    # cache
    caches_conf['tileset_cache']['cache']['type'] = tileset.cache_type

    if tileset.cache_type == 'file':
        caches_conf['tileset_cache']['cache']['directory'] = os.path.join(tileset.directory, tileset.name)
        caches_conf['tileset_cache']['cache']['directory_layout'] = tileset.directory_layout
    if tileset.cache_type == 'gpkg':
        caches_conf['tileset_cache']['cache']['filename'] = tileset.filename
        caches_conf['tileset_cache']['cache']['table_name'] = tileset.table_name

    # source
    server_service_type = tileset.server_service_type

    sources_conf['tileset_source']['type'] = u_to_str(server_service_type)

    if server_service_type == 'mapnik':
        sources_conf['tileset_source']['mapfile'] = tileset.mapnik_conf 
        sources_conf['tileset_source']['layers'] = [u_to_str(tileset.name)]
        sources_conf['tileset_source']['transparent'] = True
        sources_conf['tileset_source']['coverage'] = {}
        sources_conf['tileset_source']['coverage']['bbox'] = bbox_to_3857(tileset)
        sources_conf['tileset_source']['coverage']['srs'] = 'EPSG:3857'

    elif server_service_type == 'wms':
        sources_conf['tileset_source']['req'] = {}
        sources_conf['tileset_source']['req']['url'] = u_to_str(tileset.server_url)
        sources_conf['tileset_source']['req']['layers'] = u_to_str(tileset.layer_name)
        sources_conf['tileset_source']['req']['transparent'] = 'true'

    elif server_service_type == 'tile':
        sources_conf['sources']['tileset_source']['url'] = u_to_str(tileset.server_url)

    if tileset.server_username and tileset.server_password:
        encoded = base64.b64encode('{}:{}'.format(tileset.server_username, tileset.server_password))
        sources_conf['tileset_source']['http'] = {}
        sources_conf['tileset_source']['http']['headers'] = {}
        sources_conf['tileset_source']['http']['headers']['Authorization'] = 'Basic {}'.format(encoded)
        sources_conf['tileset_source']['http']['ssl_no_cert_checks'] = True

    # layers
    layers_conf[0]['name'] = u_to_str(tileset.name)
    layers_conf[0]['title'] = u_to_str(tileset.name)

    return json.dumps({
            'services': services_conf,
            'layers': layers_conf,
            'caches': caches_conf,
            'sources': sources_conf,
            'grids': grids_conf,
            'globals': {
                "image": {
                    "paletted": False
                }
            }
        })

def get_seed_conf(tileset):

    seed_conf = {
        'coverages': seed_coverages_conf,
        'seeds': seed_seeds_conf
    }

    if tileset.layer_zoom_start > tileset.layer_zoom_stop:
        raise ConfigurationError('invalid configuration - zoom start is greater than zoom stop')
    seed_conf['seeds']['tileset_seed']['levels']['from'] = tileset.layer_zoom_start
    seed_conf['seeds']['tileset_seed']['levels']['to'] = tileset.layer_zoom_stop

    bbox = bbox_to_3857(tileset)

    seed_conf['coverages']['tileset_geom']['bbox'] = bbox
    
    return json.dumps(seed_conf)

def bbox_to_3857(tileset):
    inProj = Proj(init='epsg:4326')
    outProj = Proj(init='epsg:3857')

    sw = transform(inProj, outProj, tileset.bbox_x0, tileset.bbox_y0)
    ne = transform(inProj, outProj, tileset.bbox_x1, tileset.bbox_y1)

    return [sw[0], sw[1], ne[0], ne[1]]

def u_to_str(string):
    return string.encode('ascii', 'ignore')
